#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward" <cward@redhat.com>

import logging

DEFAULT_CONFIG_DIR = '~/.metrique'
MONGODB_HOST = '127.0.0.1'
METRIQUE_HTTP_HOST = '127.0.0.1'
METRIQUE_HTTP_PORT = 8080

METRIQUE_CONF = 'metrique_config'
MONGODB_CONF = 'mongodb_config'
SERVER_CONFIG_PATH = DEFAULT_CONFIG_DIR
PID_FILE = '%s/server.pid' % DEFAULT_CONFIG_DIR
SSL_CERT_KEY = '%s/.metrique/pkey.pem' % DEFAULT_CONFIG_DIR
SSL_CERT = '%s/.metrique/cert.pem' % DEFAULT_CONFIG_DIR

VALID_PERMISSIONS = (None, 'r', 'rw', 'admin')

METRIQUE_DB = 'metrique'
TIMELINE_DB = 'timeline'
ADMIN_DB = 'admin'

ADMIN_USER = 'admin'
DATA_USER = 'metrique'

ETL_ACTIVITY_COLLECTION = 'etl_activity'
JOB_ACTIVITY_COLLECTION = 'job_activity'
AUTH_KEYS_COLLECTION = 'auth_keys'

CLIENTS_DB = 'clients'

LOGS_DB = 'logs'
LOGS_COLLECTION = 'logs'
MAX_SIZE = 1000000000
MAX_DOCS = 100000

DATE_FORMAT = '%Y%m%dT%H:%M:%S'
LOG_FORMAT = u'%(processName)s:%(message)s'
LOG_FORMATTER = logging.Formatter(LOG_FORMAT,
                                  DATE_FORMAT)

LOGDIR_SAVEAS = '%s/logs/metriqued.log' % DEFAULT_CONFIG_DIR
BACKUP_COUNT = 5
MAX_BYTES = 1.049e+7



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward" <cward@redhat.com>

'''
Server package covers server side of metrique,
including http api (via tornado) server side
configuration, ETL, warehouse, query,
usermanagement, logging, etc.
'''



#!/usr/bin/env python
# Author:  Jan Grec <jgrec@redhat.com>

import os
from unittest import TestCase, main

from metriqued.mongodb.basemongodb import BaseMongoDB

from jsonconf import JSONConf

config_dir = '~/.metrique/'
config_dir = os.path.expanduser(config_dir)
config = JSONConfig('mongodb', config_dir)

host = config['host']
username = config['admin_username']
password = config['admin_password']
admin_db = config['admin_db']


class Test_MongoDB_Admin(TestCase):

    def setUp(self):
        """Sets up admindb connection into test collection."""
        db = 'test'
        self.admindb = BaseMongoDB(host, username, password, db, admin_db)
        self.test = self.admindb[db]

    def test_insert(self):
        """Tests database insert function."""
        self.test.insert({'_id': 'test_insert', 'value': 'test'}, safe=True)
        self.assertTrue(self.test.find_one({'_id': 'test_insert'}))
        self.test.remove({'_id': 'test_insert'})

    def test_save(self):
        """Tests database save function."""
        self.test.save({'_id': 'test_save', 'value': 'test'}, safe=True)
        self.assertTrue(self.test.find_one({'_id': 'test_save'}))
        self.test.remove({'_id': 'test_save'})

    def test_find(self):
        """
        Tests database find function.

        At first, tries to find any data in test database. Then runs
        through returned cursor, removing each record. At the end tries
        to find data again, raising ValueError if finding is successful.
        """
        cursor = self.test.find()
        self.assertTrue(cursor)
        for record in cursor:
            self.test.remove({'_id': record['_id']})
        cursor = self.test.find()
        self.assertFalse(cursor.count(), "Find has found data in empty database")

    def test_update(self):
        """
        Tests database update function.

        Creates two records, updates one at a time, and then
        updates both at the same time using multi=True.
        """
        self.test.save({'_id': 1, 'value': 'test'})
        self.test.save({'_id': 2, 'value': 'test'})

        ud1 = {'$set': {'value': 'new_value_1', 'ad_value': 'test'}}
        ud2 = {'$set': {'value': 'new_value_2', 'ad_value': 'test'}}
        self.test.update({'_id': 1}, ud1)
        self.test.update({'_id': 2}, ud2)

        new_record_1 = self.test.find_one({'_id': 1})
        self.assertEqual(new_record_1['value'], u'new_value_1')
        new_record_2 = self.test.find_one({'_id': 2})
        self.assertEqual(new_record_2['value'], u'new_value_2')

        nud = {'$set': {'value': 'the_newest_value', 'ad_value': 'changed'}}
        self.test.update({'_id': {'$gt': 0}}, nud, multi=True)

        for record in self.test.find():
            self.assertEqual(record['value'], u'the_newest_value')
            self.assertEqual(record['ad_value'], u'changed')

        self.test.remove({'_id': 1})
        self.test.remove({'_id': 2})

    def test_ensure_index(self):
        """
        Tests database ensure_index function.

        Creates three records with ascending integer '_id' and descending
        character 'value'. Then creates an unnamed ascending index on 'value'
        field.
        """
        self.test.drop_indexes()
        self.test.save({'_id': 1, 'value': 'c'})
        self.test.save({'_id': 2, 'value': 'b'})
        self.test.save({'_id': 3, 'value': 'a'})
        key = str("value")
        self.test.ensure_index(key, 5, name="testingdex")
        cursor = self.test.find({'value': {'$gt': 'b'}})
        for record in cursor:
            self.assertTrue(record['_id'], 1)
        self.test.drop_indexes()
        self.test.remove({'_id': 1})
        self.test.remove({'_id': 2})
        self.test.remove({'_id': 3})


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
import os

from defaults import LOGDIR_SAVEAS
from defaults import ADMIN_DB, METRIQUE_DB
from defaults import JOB_ACTIVITY_COLLECTION
from defaults import LOGS_COLLECTION, AUTH_KEYS_COLLECTION
from defaults import LOG_FORMATTER
from defaults import ETL_ACTIVITY_COLLECTION, MAX_SIZE, MAX_DOCS
from defaults import DATA_USER, ADMIN_USER, TIMELINE_DB
from defaults import PID_FILE
from defaults import SSL_CERT, SSL_CERT_KEY
from defaults import MONGODB_CONF
from defaults import DEFAULT_CONFIG_DIR
from defaults import MONGODB_HOST
from defaults import METRIQUE_HTTP_HOST
from defaults import METRIQUE_HTTP_PORT

from jsonconf import JSONConf

from metriqued.mongodb.basemongodb import BaseMongoDB


class metrique(JSONConf):
    def __init__(self, config_file, config_dir=None,
                 force=True, *args, **kwargs):
        if not config_dir:
            config_dir = DEFAULT_CONFIG_DIR
        super(metrique, self).__init__(config_file, config_dir,
                                       force=force, *args, **kwargs)
        self._properties = {}

    @property
    def pid_file(self):
        _pf = os.path.expanduser(PID_FILE)
        return self._default('pid_file', _pf)

    @property
    def server_thread_count(self):
        from multiprocessing import cpu_count
        return self._default('server_thread_count', cpu_count() * 10)

    @property
    def async(self):
        return self._default('async', True)

    @async.setter
    def async(self, bool_):
        self.config['async'] = bool_

    @property
    def debug(self):
        ''' Reflect whether debug is enabled or not '''
        return self._default('debug', -1)

    @debug.setter
    def debug(self, n):
        logger = logging.getLogger('metrique')
        if n == -1:
            level = logging.WARN
        elif n == 0:
            level = logging.INFO
        elif n == 1:
            level = logging.DEBUG
            logger.debug("Debug: metrique")
        elif n == 2:
            logger = logging.getLogger()
            level = logging.DEBUG
            logger.debug("Debug: metrique, tornado")
        logger.setLevel(level)
        self.config['debug'] = n

    @property
    def admin_user(self):
        return self._default('admin_user', ADMIN_USER)

    @property
    def admin_password(self):
        return self._default('admin_password', None)

    @admin_password.setter
    def admin_password(self, value):
        self.config['admin_password'] = value

    @property
    def auth(self):
        return self._default('auth', False)

    @auth.setter
    def auth(self, value):
        self.config['auth'] = value

    @property
    def gzip(self):
        return self._default('gzip', True)

    @gzip.setter
    def gzip(self, value):
        self.config['gzip'] = value

    @property
    def http_host(self):
        return self._default('http_host', METRIQUE_HTTP_HOST)

    @http_host.setter
    def http_host(self, value):
        self.config['http_host'] = value

    @property
    def http_port(self):
        return self._default('http_port', METRIQUE_HTTP_PORT)

    @http_port.setter
    def http_port(self, value):
        self.config['http_port'] = value

    @property
    def krb_realm(self):
        return self._default('krb_realm', '')

    @property
    def static_path(self):
        abspath = os.path.dirname(os.path.abspath(__file__))
        static_path = os.path.join(abspath, 'static/')
        return self._default('static_path', static_path)

    @property
    def ssl(self):
        return self._default('ssl', False)

    @ssl.setter
    def ssl(self, value):
        self.config['ssl'] = value

    @property
    def ssl_certificate_key(self):
        return self._default(
            'ssl_certificate_key', os.path.expanduser(SSL_CERT_KEY))

    @ssl_certificate_key.setter
    def ssl_certificate_key(self, value):
        self.config['ssl_certificate_key'] = value

    @property
    def ssl_certificate(self):
        return self._default('ssl_certificate', os.path.expanduser(SSL_CERT))

    @ssl_certificate.setter
    def ssl_certificate(self, value):
        self.config['ssl_certificate'] = value

    @property
    def logs_max_size(self):
        return self._default('logs_max_size', MAX_SIZE)

    @property
    def logs_max_docs(self):
        return self._default('logs_max_docs', MAX_DOCS)

    @property
    def log_formatter(self):
        return self._property_default('log_formatter', LOG_FORMATTER)

    @property
    def log_to_stdout(self):
        return self._default('log_to_stdout', 1)

    @property
    def log_file_path(self):
        return os.path.expanduser(
            self._default('log_file_path', LOGDIR_SAVEAS))

    @log_file_path.setter
    def log_file_path(self, value):
        self.config['log_file_path'] = value

    @property
    def log_to_file(self):
        return self._default('log_to_file', 0)


class mongodb(JSONConf):
    def __init__(self, config_file=None, config_dir=None,
                 force=True, *args, **kwargs):
        if not config_file:
            config_file = MONGODB_CONF
        if not config_dir:
            config_dir = DEFAULT_CONFIG_DIR
        super(mongodb, self).__init__(config_file, config_dir,
                                      force=force, *args, **kwargs)

    @property
    def host(self):
        return self._default('host', MONGODB_HOST)

    @host.setter
    def host(self, value):
        self.config['host'] = value

    @property
    def ssl(self):
        return self._default('ssl', False)

    @host.setter
    def host(self, value):
        self.config['host'] = value

    @property
    def admin_password(self):
        return self._default('admin_password', None)

    @admin_password.setter
    def admin_password(self, value):
        self.config['admin_password'] = value

    @property
    def data_password(self):
        return self._default('data_password', None)

    @data_password.setter
    def data_password(self, value):
        self.config['data_password'] = value

    @property
    def data_user(self):
        return self._default('data_user', DATA_USER)

    @property
    def admin_user(self):
        return self._default('admin_user', ADMIN_USER)

    @property
    def db_admin(self):
        return self._default('db_admin', ADMIN_DB)

    @property
    def db_metrique(self):
        return self._default('db_metrique', METRIQUE_DB)

    @property
    def db_timeline(self):
        return self._default('db_timeline', TIMELINE_DB)

    @property
    def db_metrique_admin(self):
        return BaseMongoDB(host=self.host, db=self.db_metrique,
                           user=self.admin_user,
                           password=self.admin_password,
                           admin_db=self.db_admin,
                           ssl=self.ssl)

    @property
    def db_timeline_admin(self):
        return BaseMongoDB(host=self.host, db=self.db_timeline,
                           user=self.admin_user,
                           password=self.admin_password,
                           admin_db=self.db_admin,
                           ssl=self.ssl)

    @property
    def db_timeline_data(self):
        return BaseMongoDB(host=self.host, db=self.db_timeline,
                           user=self.data_user,
                           password=self.data_password,
                           ssl=self.ssl)

    @property
    def collection_etl(self):
        return self._default('collection_etl', ETL_ACTIVITY_COLLECTION)

    @property
    def collection_jobs(self):
        return self._default('collection_jobs', JOB_ACTIVITY_COLLECTION)

    @property
    def collection_logs(self):
        return self._default('collection_logs', LOGS_COLLECTION)

    @property
    def collection_auth_keys(self):
        return self._default('collection_auth_keys', AUTH_KEYS_COLLECTION)

    @property
    def c_job_activity(self):
        return self.db_metrique_admin[self.collection_jobs]

    @property
    def c_logs(self):
        return self.db_metrique_admin[self.collection_logs]

    @property
    def c_etl_activity(self):
        return self.db_metrique_admin[self.collection_etl]

    @property
    def c_auth_keys(self):
        return self.db_metrique_admin[self.collection_auth_keys]



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
import base64
logger = logging.getLogger(__name__)

from functools import wraps
import kerberos
from passlib.hash import sha256_crypt
import simplejson as json
import tornado
import traceback

from metriqued.defaults import VALID_PERMISSIONS as VP
from metriqued import query_api, etl_api, users_api

__all__ = '__all__'


def async(f):
    '''
    Decorator for enabling async Tornado.Handlers
    If not metrique.config.async: disable async

    Requires: futures

    Uses: async.threading
    '''
    @tornado.web.asynchronous
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.proxy.metrique_config.async:
            def future_end(future):
                try:
                    _result = future.result()
                    logger.debug('JSON dump: START... ')
                    result = json.dumps(_result, ensure_ascii=False)
                    logger.debug('JSON dump: DONE')
                except Exception:
                    result = traceback.format_exc()
                    logger.error(result)
                    raise tornado.web.HTTPError(500, result)
                finally:
                    self.write(result)
                self.finish()

            future = self.proxy.executor.submit(f, self, *args, **kwargs)
            tornado.ioloop.IOLoop.instance().add_future(future, future_end)
        else:
            _result = f(self, *args, **kwargs)
            logger.debug('JSON dump: START... ')
            result = json.dumps(_result, ensure_ascii=False)
            logger.debug('JSON dump: DONE... ')
            self.write(result)
            self.finish()
    return wrapper


def request_authentication(handler):
    ''' Helper-Function for settig 401 - Request for authentication '''
    handler.set_status(401)
    handler.set_header('WWW-Authenticate', 'Basic realm="Metrique"')


def _auth_admin(handler, username, password):
    '''
    admin pass is stored in metrique server config
    admin user gets 'rw' to all cubes
    '''
    admin_user = handler.proxy.metrique_config.admin_user
    admin_pass = handler.proxy.metrique_config.admin_password
    if username == admin_user:
        if password == admin_pass:
            return True
        else:
            return -1
    else:
        return 0


def _auth_kerb(handler, username, password):
    if not password:
        return -1
    krb_realm = handler.proxy.metrique_config.krb_realm
    if not krb_realm:
        return 0

    try:
        ret = kerberos.checkPassword(username,
                                     password, '',
                                     krb_realm)
        return ret
    except kerberos.BasicAuthError as e:
        logger.debug('KRB ERROR: %s' % e)
        return -1


def _auth_basic(handler, password, user_dict):
    if not (password and user_dict and user_dict.get('password')):
        # only authenticate if we actually have a password
        # and user_dict.password to compare
        return -1
    else:
        p_hash = user_dict.get('password')
        return sha256_crypt.verify(password, p_hash)


def _get_resource_acl(handler, resource, lookup):
    ''' Check if user is listed to access to a given resource '''
    resource = [resource, __all__]
    _lookup = [{lookup:  {'$exists': True}},
               {__all__: {'$exists': True}}]
    spec = {'_id': {'$in': resource},
            '$or': _lookup}
    logger.debug("Cube Check: spec (%s)" % spec)
    return handler.proxy.mongodb_config.c_auth_keys.find_one(spec)


def _check_acl(permissions, user):
    try:
        return VP.index(user['permissions']) >= VP.index(permissions)
    except (TypeError, KeyError):
        # permissions is not defined; assume they're unpermitted
        return -1


def _check_perms(doc, username, cube, permissions):
    if doc and '__all__' in doc:
        user = doc['__all__']
        is_ok = _check_acl(permissions, user)
        no_auth = True
    elif doc:
        user = doc[username]
        is_ok = _check_acl(permissions, user)
        no_auth = False
    elif cube:
        user = {}
        is_ok = -1
        no_auth = False
    return user, is_ok, no_auth


def authenticate(handler, username, password, permissions):
    ''' Helper-Function for determining whether a given
        user:password:permissions combination provides
        client with enough privleges to execute
        the requested command against the given cube
    '''
    cube = handler.get_argument('cube')

    is_admin = _auth_admin(handler, username, password)
    if is_admin in [True, -1]:
        # ... or if user is admin with correct admin pass
        logger.debug('AUTH OK: admin')
        return is_admin

    doc = _get_resource_acl(handler, cube, username)
    user, is_ok, no_auth = _check_perms(doc, username, cube, permissions)

    if not is_ok:
        return is_ok
    elif no_auth:
        return True
    elif _auth_kerb(handler, username, password) is True:
        # or if user is kerberous auth'd
        logger.debug(
            'AUTH OK: krb (%s:%s)' % (username, cube))
        return True
    elif _auth_basic(handler, password, user) is True:
        # or if the user is authed by metrique (built-in; auth_keys)
        logger.debug(
            'AUTH OK: basic (%s:%s)' % (username, cube))
        return True
    else:
        return -1


def auth(permissions='r'):
    ''' Decorator for auth dependent Tornado.Handlers '''
    # permissions must be a single string
    assert isinstance(permissions, basestring)
    assert permissions in VP

    def decorator(f):
        @wraps(f)
        def wrapper(handler, *args, **kwargs):
            if not handler.proxy.metrique_config.auth:
                # if auth isn't on, let anyone do anything!
                return f(handler, *args, **kwargs)

            auth_header = handler.request.headers.get('Authorization')
            if auth_header is None or not auth_header.startswith('Basic '):
                # No HTTP Basic Authentication header
                return request_authentication(handler)

            auth = base64.decodestring(auth_header[6:])
            username, password = auth.split(':', 2)

            privleged = authenticate(handler, username, password, permissions)
            logger.debug("User (%s): Privleged (%s)" % (username, privleged))
            if privleged in [True, 1]:
                return f(handler, *args, **kwargs)
            else:
                raise tornado.web.HTTPError(401)
        return wrapper
    return decorator


class MetriqueInitialized(tornado.web.RequestHandler):
    '''
        Template RequestHandler that accepts init parameters
        and unifies json get_argument handling
    '''

    def initialize(self, proxy):
        '''
        Paremeters
        ----------
        proxy : HTTPServer (MetriqueServer) Obj
            A pointer to the running metrique server namespace
        '''
        self.proxy = proxy

    def get_argument(self, key, default=None):
        '''
            Assume incoming arguments are json encoded,
            get_arguments should always deserialize
            on the way in
        '''
        # arguments are expected to be json encoded!
        _arg = super(MetriqueInitialized, self).get_argument(key, default)

        if _arg is None:
            return _arg

        try:
            arg = json.loads(_arg)
        except Exception as e:
            raise ValueError("Invalid JSON content (%s): %s" % (type(_arg), e))
        return arg


class PingHandler(MetriqueInitialized):
    ''' RequestHandler for pings'''
    @async
    def get(self):
        return self.proxy.ping()


class QueryAggregateHandler(MetriqueInitialized):
    '''
        RequestHandler for running mongodb aggregation
        framwork pipeines against a given cube
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        pipeline = self.get_argument('pipeline', '[]')
        return query_api.aggregate(cube, pipeline)


class QueryFetchHandler(MetriqueInitialized):
    ''' RequestHandler for fetching lumps of cube data '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        fields = self.get_argument('fields')
        date = self.get_argument('date')
        sort = self.get_argument('sort', None)
        skip = self.get_argument('skip', 0)
        limit = self.get_argument('limit', 0)
        oids = self.get_argument('oids', [])
        return query_api.fetch(cube=cube, fields=fields, date=date,
                               sort=sort, skip=skip, limit=limit, oids=oids)


class QueryCountHandler(MetriqueInitialized):
    '''
        RequestHandler for returning back simple integer
        counts of objects matching the given query
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        query = self.get_argument('query')
        date = self.get_argument('date', None)
        return query_api.count(cube=cube, query=query, date=date)


class QueryFindHandler(MetriqueInitialized):
    '''
        RequestHandler for returning back object
        matching the given query
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        query = self.get_argument('query')
        fields = self.get_argument('fields', '')
        date = self.get_argument('date')
        sort = self.get_argument('sort', None)
        one = self.get_argument('one', False)
        explain = self.get_argument('explain', False)
        return query_api.find(cube=cube,
                              query=query,
                              fields=fields,
                              date=date,
                              sort=sort,
                              one=one,
                              explain=explain)


class QueryDeptreeHandler(MetriqueInitialized):
    '''
        RequestHandler for returning back the list of
        oids matching the given tree.
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        field = self.get_argument('field')
        oids = self.get_argument('oids')
        date = self.get_argument('date')
        level = self.get_argument('level')
        return query_api.deptree(cube=cube,
                                 field=field,
                                 oids=oids,
                                 date=date,
                                 level=level)


class QueryDistinctHandler(MetriqueInitialized):
    '''
        RequestHandler for fetching distinct token values for a
        given cube.field
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        field = self.get_argument('field')
        return query_api.distinct(cube=cube, field=field)


class UsersAddHandler(MetriqueInitialized):
    '''
        RequestHandler for managing user access control
        lists for a given cube
    '''
    @auth('admin')
    @async
    def get(self):
        cube = self.get_argument('cube')
        user = self.get_argument('user')
        password = self.get_argument('password')
        permissions = self.get_argument('permissions', 'r')
        return users_api.add(user, password, permissions, cube)


class ETLIndexHandler(MetriqueInitialized):
    '''
        RequestHandler for ensuring mongodb indexes
        in timeline collection for a given cube
    '''
    @auth('rw')
    @async
    def get(self):
        cube = self.get_argument('cube')
        ensure = self.get_argument('ensure')
        drop = self.get_argument('drop')
        return etl_api.index(cube=cube, ensure=ensure, drop=drop)


class ETLActivityImportHandler(MetriqueInitialized):
    '''
        RequestHandler for building pre-calculated
        object timelines given a 'activity history'
        data source that can be used to recreate
        objects in time
    '''
    @auth('rw')
    @async
    def get(self):
        cube = self.get_argument('cube')
        ids = self.get_argument('ids')
        return etl_api.activity_import(cube=cube, ids=ids)


class ETLSaveObjects(MetriqueInitialized):
    '''
        RequestHandler for saving a given
        object to a metrique server cube
    '''
    @auth('rw')
    @async
    def post(self):
        cube = self.get_argument('cube')
        objects = self.get_argument('objects')
        update = self.get_argument('update')
        mtime = self.get_argument('mtime')
        return etl_api.save_objects(cube=cube, objects=objects, update=update,
                                    mtime=mtime)


class ETLRemoveObjects(MetriqueInitialized):
    '''
        RequestHandler for saving a given
        object to a metrique server cube
    '''
    @auth('rw')
    @async
    def delete(self):
        cube = self.get_argument('cube')
        ids = self.get_argument('ids')
        backup = self.get_argument('backup')
        return etl_api.remove_objects(cube=cube, ids=ids,
                                      backup=backup)


class ETLCubeDrop(MetriqueInitialized):
    ''' RequestsHandler for droping given cube from timeline '''
    @auth('rw')
    @async
    def delete(self):
        cube = self.get_argument('cube')
        return etl_api.drop(cube=cube)


class CubeHandler(MetriqueInitialized):
    '''
        RequestHandler for querying about
        available cubes and cube.fields
    '''
    @auth('r')
    @async
    def get(self):
        cube = self.get_argument('cube')
        _mtime = self.get_argument('_mtime')
        exclude_fields = self.get_argument('exclude_fields')
        if cube is None:
            # return a list of cubes
            return self.proxy.list_cubes()
        else:
            # return a list of fields in a cube
            return self.proxy.list_cube_fields(cube,
                                               exclude_fields,
                                               _mtime)



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
logger = logging.getLogger(__name__)

import os
import tornado.ioloop
import tornado.web

from metriqued.metriqueserver import MetriqueServer

from handlers import PingHandler
from handlers import QueryAggregateHandler, QueryFindHandler
from handlers import QueryDeptreeHandler
from handlers import QueryFetchHandler, QueryCountHandler
from handlers import QueryDistinctHandler
from handlers import UsersAddHandler
from handlers import ETLIndexHandler
from handlers import ETLActivityImportHandler
from handlers import ETLSaveObjects, ETLRemoveObjects, ETLCubeDrop
from handlers import CubeHandler


class HTTPServer(MetriqueServer):
    ''' HTTP (Tornado >=3.0) implemntation of MetriqueServer '''
    def __init__(self, host=None, port=None, **kwargs):
        super(HTTPServer, self).__init__(**kwargs)
        if host:
            self.metrique_config.http_host = host
        if port:
            self.metrique_config.http_port = port

    def _setup_webapp(self):
        ''' Config and Views'''
        logger.debug("Tornado: Web App setup")
        init = dict(proxy=self)
        debug = self.metrique_config.debug == 2
        gzip = self.metrique_config.gzip
        static_path = self.metrique_config.static_path
        self._web_app = tornado.web.Application(
            gzip=gzip,
            debug=debug,
            static_path=static_path,
            handlers=[
                (r"/api/v1/ping/?", PingHandler, init),
                (r"/api/v1/query/find", QueryFindHandler, init),
                (r"/api/v1/query/deptree", QueryDeptreeHandler, init),
                (r"/api/v1/query/count", QueryCountHandler, init),
                (r"/api/v1/query/aggregate", QueryAggregateHandler, init),
                (r"/api/v1/query/fetch", QueryFetchHandler, init),
                (r"/api/v1/query/distinct", QueryDistinctHandler, init),
                (r"/api/v1/admin/users/add", UsersAddHandler, init),
                (r"/api/v1/admin/etl/index", ETLIndexHandler, init),
                (r"/api/v1/admin/etl/activityimport",
                    ETLActivityImportHandler, init),
                (r"/api/v1/admin/etl/saveobjects", ETLSaveObjects, init),
                (r"/api/v1/admin/etl/removeobjects", ETLRemoveObjects, init),
                (r"/api/v1/admin/etl/cube/drop", ETLCubeDrop, init),
                (r"/api/v1/cube", CubeHandler, init),
            ],
        )
        port = self.metrique_config.http_port
        address = self.metrique_config.http_host
        if self.metrique_config.ssl:
            ssl_options = dict(
                certfile=self.metrique_config.ssl_certificate,
                keyfile=self.metrique_config.ssl_certificate_key)
            try:
                self._web_app.listen(port=port, address=address,
                                     ssl_options=ssl_options)
            except ValueError:
                raise ValueError(
                    "SSL Cert missing, perhaps? (%s)." % ssl_options)
        else:
            self._web_app.listen(port=port, address=address)

        logger.debug("Tornado: listening on %s:%s" % (port, address))

    def start(self):
        ''' Start a new tornado web app '''
        pid = os.fork()
        if pid == 0:
            logger.debug("Tornado: Start")
            super(HTTPServer, self).start()
            self._setup_webapp()
            ioloop = tornado.ioloop.IOLoop.instance()
            ioloop.start()

    def stop(self):
        ''' Stop a run tornado web app '''
        # FIXME: THIS ISN"T WORKING!
        # should catch sigkill/sigterm and shutdown properly
        super(HTTPServer, self).stop()
        logger.debug("Tornado: Stop")
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.stop()
        if hasattr(self, '_web_app'):
            del self._web_app



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
logging.basicConfig()
logger = logging.getLogger(__name__)

from concurrent.futures import ThreadPoolExecutor
import os

from metriqued.baseserver import BaseServer
from metriqued.defaults import BACKUP_COUNT, MAX_BYTES


class MetriqueServer(BaseServer):
    def __init__(self, **kwargs):
        super(MetriqueServer, self).__init__(**kwargs)
        logger.debug('Debug: %s' % self.metrique_config.debug)
        logger.debug('Async: %s' % self.metrique_config.async)
        logger.debug(' Auth: %s' % self.metrique_config.auth)
        logger.debug('  SSL: %s' % self.metrique_config.ssl)

    @property
    def pid(self):
        if os.path.exists(self.metrique_config.pid_file):
            lines = open(self.metrique_config.pid_file).readlines()
            if not lines:
                pid = 0
            else:
                pid = int(lines[0])
        else:
            pid = 0
        return pid

    def _set_pid(self):
        if self.pid:
            raise RuntimeError(
                "(%s) found in %s" % (self.pid, self.metrique_config.pid_file))

        _pid = os.getpid()
        with open(self.metrique_config.pid_file, 'w') as file:
            file.write(str(_pid))
        return _pid

    def _remove_pid(self):
        try:
            os.remove(self.metrique_config.pid_file)
        except IOError:
            pass

    def start(self):
        if self.metrique_config.log_to_file:
            hdlr = logging.handlers.RotatingFileHandler(
                self.log_file_path, maxBytes=MAX_BYTES,
                backupCount=BACKUP_COUNT)
            hdlr.setFormatter(self.metrique_config.log_formatter)
            logger.addHandler(hdlr)
        self._set_pid()
        k = self.metrique_config.server_thread_count
        self.executor = ThreadPoolExecutor(k)
        logger.debug("Metrique Server - Started")

        # Fail to start if we can't communicate with mongo
        try:
            assert self.mongodb_config.db_metrique_admin.db
        except Exception as e:
            host = self.mongodb_config.host
            raise SystemExit(
                '%s\nFailed to communicate with MongoDB (%s)' % (e, host))

    def stop(self):
        self._remove_pid()
        self.executor.shutdown()
        logger.debug("Metrique Server - Stopped")



#!/usr/bin/env python
# Author:  Jan Grec <jgrec@redhat.com>

from datetime import datetime
import os
import time
from unittest import TestCase, TestLoader, TextTestRunner

from metriqued.Utils.MongoDB.BaseMongoDB import BaseMongoDB

from jsonconf import JSONConf

config_dir = '~/.metrique/'
config_dir = os.path.expanduser(config_dir)
config = JSONConfig('mongodb', config_dir)

host = config['host']
username = config['admin_username']
password = config['admin_password']
admin_db = config['admin_db']
tests_db = 'tests'
junk_c = 'junk'
results_c = 'results'

TEST_RANGE = 10000   # How many operation iterations will be done during one test
JUNK_RANGE = TEST_RANGE + 1  # Pre-created junk collection records count
                             # MUST BE GREATER THAN TEST_RANGE
SAFE = False

# WARNING
# These tests creates new collection ('tests') in metrique database and in this
# collection creates this dbs:
#   - junk (storing testing data - removed after each test)
#   - results (storing results after each test - doesn't get removed)


class Test_MongoDB_Performance(TestCase):

    def setUp(self):
        """Opens connection and creates 'junk' records."""
        self.admindb = BaseMongoDB(host, username, password,
                                   tests_db, admin_db)
        self.junk = self.admindb[junk_c]
        self.results = self.admindb[results_c]
        self.junk.remove(safe=SAFE)
        for record in range(0, JUNK_RANGE):
            self.junk.insert({'_id': record, 'value': 'test_value'}, safe=SAFE)

    def tearDown(self):
        """Removes all 'junk' records."""
        self.junk.remove(safe=SAFE)
        self.admindb.close()
        del(self.junk)
        del(self.results)
        del(self.admindb)

    def save_result(self, case, time):
        """Stores results of performance tests in mongo database."""
        result = {'date': datetime.utcnow(), 'test': 'MongoDB-performance',
                  'case': case, 'result': time, 'range': TEST_RANGE}
        self.results.insert(result)

    def test_remove_ascending(self):
        """Remove test - ascending."""
        id_set = range(0, TEST_RANGE)
        S = time.time()
        for sid in id_set:
            self.junk.remove({'_id': sid}, safe=SAFE)
        D = time.time() - S
        self.save_result('remove (asc)', D)

    def test_find_ascending(self):
        """Find test - ascending."""
        id_set = range(0, TEST_RANGE)
        S = time.time()
        for sid in id_set:
            self.junk.find({'_id': sid})
        D = time.time() - S
        self.save_result('find (asc)', D)

    def test_findone_ascending(self):
        """Find_one test - ascending."""
        id_set = range(0, TEST_RANGE)
        S = time.time()
        for sid in id_set:
            self.junk.find_one({'_id': sid})
        D = time.time() - S
        self.save_result('find_one (asc)', D)

    def test_save(self):
        """Save test."""
        id_set = range(0, TEST_RANGE)
        S = time.time()
        for sid in id_set:
            record = {'_id': sid, 'value': 'test_junks'}
            self.junk.save(record, safe=SAFE)
        D = time.time() - S
        self.save_result('save', D)

    def test_updatemulti(self):
        """Update test (multi=True)."""
        record = {'$set': {'value': 'test_junks'}}
        S = time.time()
        self.junk.update({'_id': {'$lt': TEST_RANGE}}, record, multi=True, safe=SAFE)
        D = time.time() - S
        self.save_result('update-multi', D)

    def test_update_ascending(self):
        """Update test - ascending."""
        id_set = range(0, TEST_RANGE)
        record = {'$set': {'value': 'test_junks'}}
        S = time.time()
        for sid in id_set:
            self.junk.update({'_id': sid}, record, safe=SAFE)
        D = time.time() - S
        self.save_result('update (asc)', D)

    def test_insert(self):
        """Insert test."""
        id_set = range(JUNK_RANGE, JUNK_RANGE + TEST_RANGE)
        S = time.time()
        for sid in id_set:
            record = {'_id': sid, 'value': 'test_dumps'}
            self.junk.insert(record, safe=SAFE)
        D = time.time() - S
        self.save_result('insert', D)


def print_results():
    admindb = BaseMongoDB(host, username, password,
                          tests_db, admin_db)
    results = admindb[results_c]

    res_table = {}

    cases = results.distinct('case')
    for case in cases:

        cur = results.find({"case": case}).sort("date", -1)
        try:
            actual_rec = cur.next()
            actual = actual_rec[u'result']
            actual_average = actual / actual_rec[u'range']
        except StopIteration:
            actual = 0.0
            actual_average = 0.0
            previous = 0.0
            previous_average = 0.0
        try:
            previous_rec = cur.next()
            previous = previous_rec[u'result']
            previous_average = previous / previous_rec[u'range']
        except StopIteration:
            previous = 0.0
            previous_average = 0.0

        cur = results.find({"case": case}).sort("result", 1)
        try:
            best_rec = cur.next()
            best = best_rec[u'result']
            best_average = best / best_rec[u'range']
        except StopIteration:
            best = 0.0
            best_average = 0.0

        cur = results.find({"case": case}).sort("result", -1)
        try:
            worst_rec = cur.next()
            worst = worst_rec[u'result']
            worst_average = worst / worst_rec[u'range']
        except:
            worst = 0.0
            worst_average = 0.0

        res_table[case] = {'actual': actual, 'actual_average': actual_average,
                           'previous': previous, 'previous_average': previous_average,
                           'best': best, 'best_average': best_average,
                           'worst': worst, 'worst_average': worst_average}

    admindb.close()

    print "name" + (14 * ' ') + "actual  act(av)   prev    prev(av)  best    best(av)  worst   worst(av)"
    print (89 * '-')
    for case in cases:
        spaces = 18 - len(str(case))
        table = "%3.4f  %3.6f  %3.4f  %3.6f  %3.4f  %3.6f  %3.4f  %3.6f" % (res_table[case]['actual'],
                                                                            res_table[case]['actual_average'],
                                                                            res_table[case]['previous'],
                                                                            res_table[case]['previous_average'],
                                                                            res_table[case]['best'],
                                                                            res_table[case]['best_average'],
                                                                            res_table[case]['worst'],
                                                                            res_table[case]['worst_average'])
        print str(case) + (spaces * ' ') + table

if __name__ == '__main__':
    suite = TestLoader().loadTestsFromTestCase(Test_MongoDB_Performance)
    TextTestRunner(verbosity=2).run(suite)
    print_results()



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
logger = logging.getLogger(__name__)
import pql
import re

from metriqued.cubes import get_fields, get_cube
from metriqued.job import job_save
from metriqued.utils import dt2ts

BATCH_SIZE = 16777216  # hard limit is 16M...


@job_save('query distinct')
def distinct(cube, field):
    logger.debug('Running Distinct (%s.%s)' % (cube, field))
    _cube = get_cube(cube)
    return _cube.distinct(field)


@job_save('query count')
def count(cube, query, date=None):
    logger.debug('Running Count')
    try:
        spec = pql.find(query + _get_date_pql_string(date))
    except Exception as e:
        raise ValueError("Invalid Query (%s)" % str(e))

    _cube = get_cube(cube)

    logger.debug('Mongo Query: %s' % spec)

    docs = _cube.find(spec)
    result = docs.count() if docs else 0
    docs.close()
    return result


def _get_date_pql_string(date, prefix=' and '):
    if date is None:
        return prefix + '_end == None'
    if date == '~':
        return ''

    dt_str = date.replace('T', ' ')
    dt_str = re.sub('(\+\d\d:\d\d)?$', '', dt_str)

    before = lambda d: '_start <= %f' % dt2ts(d)
    after = lambda d: '(_end >= %f or _end == None)' % dt2ts(d)
    split = date.split('~')
    logger.warn(split)
    if len(split) == 1:
        ret = '%s and %s' % (before(dt_str), after(dt_str))
    elif split[0] == '':
        ret = '%s' % before(split[1])
    elif split[1] == '':
        ret = '%s' % after(split[0])
    else:
        ret = '%s and %s' % (before(split[1]), after(split[0]))
    return prefix + ret


def _check_sort(sort):
    if not sort:
        sort = [('_oid', 1)]

    try:
        assert len(sort[0]) == 2
    except (AssertionError, IndexError, TypeError):
        raise ValueError("Invalid sort value; try [('_id': -1)]")

    return sort


@job_save('query find')
def find(cube, query, fields=None, date=None, sort=None, one=False,
         explain=False):
    logger.debug('Running Find (%s)' % cube)

    sort = _check_sort(sort)
    _cube = get_cube(cube)
    fields = get_fields(cube, fields)

    query += _get_date_pql_string(date)

    logger.debug("PQL Query: %s" % query)
    pql_parser = pql.SchemaFreeParser()
    try:
        spec = pql_parser.parse(query)
    except Exception as e:
        raise ValueError("Invalid Query (%s): %s" % (str(e), query))

    logger.debug('Query: %s' % spec)

    if explain:
        result = _cube.find(spec, fields, sort=sort).explain()
    elif one:
        result = _cube.find_one(spec, fields, sort=sort)
    else:
        result = _cube.find(spec, fields, sort=sort)
        result.batch_size(BATCH_SIZE)
        result = tuple(result)
    return result


def _parse_oids(oids, delimeter=','):
    if isinstance(oids, basestring):
        oids = [s.strip() for s in oids.split(delimeter)]
    if type(oids) is not list:
        raise TypeError("ids expected to be a list")
    return oids


@job_save('query deptree')
def deptree(cube, field, oids, date, level):
    oids = _parse_oids(oids)
    _cube = get_cube(cube)
    checked = set(oids)
    fringe = oids
    loop_k = 0

    while len(fringe) > 0:
        if level and loop_k == abs(level):
            break
        spec = pql.find('_oid in %s and %s != None' % (fringe, field) +
                        _get_date_pql_string(date))
        deps = _cube.find(spec, ['_oid', field])
        fringe = set([oid for doc in deps for oid in doc[field]])
        fringe = filter(lambda oid: oid not in checked, fringe)
        checked |= set(fringe)
        loop_k += 1
    return sorted(checked)


@job_save('query fetch')
def fetch(cube, fields=None, date=None, sort=None, skip=0, limit=0, oids=None):
    if oids is None:
        oids = []
    logger.debug('Running Fetch (skip:%s, limit:%s, oids:%s)' % (
        skip, limit, len(oids)))

    sort = _check_sort(sort)
    _cube = get_cube(cube)
    fields = get_fields(cube, fields)

    spec = {'_oid': {'$in': _parse_oids(oids)}} if oids else {}
    dt_str = _get_date_pql_string(date, '')
    if dt_str:
        spec.update(pql.find(dt_str))

    result = _cube.find(spec, fields, sort=sort, skip=skip, limit=limit)
    result.batch_size(BATCH_SIZE)
    result = tuple(result)
    return result


@job_save('query aggregate')
def aggregate(cube, pipeline):
    logger.debug('Running Aggregation')
    logger.debug('Pipeline (%s): %s' % (type(pipeline), pipeline))
    _cube = get_cube(cube)
    return _cube.aggregate(pipeline)



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward" <cward@redhat.com>

import logging
logger = logging.getLogger(__name__)

from pymongo import Connection, errors

DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 27017
DEFAULT_SSL = False
DEFAULT_TIMEOUT = None
DEFAULT_WRITE_CONCERN = 1


class BaseMongoDB(object):
    def __init__(self, db, host=DEFAULT_HOST, user=None, password=None,
                 admin_db=None, ssl=DEFAULT_SSL, port=DEFAULT_PORT,
                 timeout=DEFAULT_TIMEOUT,
                 write_concern=DEFAULT_WRITE_CONCERN):
        self._host = host
        self._password = password
        self._user = user
        self._db = db
        self._admin_db = admin_db

        self._ssl = ssl
        self._port = port
        self._timeout = timeout
        self._write_concern = write_concern

    @property
    def db(self):
        if not hasattr(self, '_db_proxy'):
            try:
                self._proxy = Connection(self._host, self._port,
                                         ssl=self._ssl,
                                         tz_aware=True,
                                         network_timeout=self._timeout)
            except errors.ConnectionFailure as e:
                raise errors.ConnectionFailure(
                    "MongoDB Failed to connect (%s): %s" % (self._host, e))
            self._db_proxy = self._auth_db()
        # return the connected, authenticated database object
        return self._db_proxy

    def close(self):
        if hasattr(self, '_proxy'):
            self._proxy.close()

    def _auth_db(self):
        '''
        by default the default user only has read-only access to db
        '''
        if not self._password:
            pass
        elif self._admin_db:
            admin_db = self._proxy[self._admin_db]
            if not admin_db.authenticate(self._user, self._password):
                raise RuntimeError(
                    "MongoDB failed to authenticate user (%s)" % self._user)
        else:
            if not self._proxy[self._db].authenticate(self._user,
                                                      self._password):
                raise RuntimeError(
                    "MongoDB failed to authenticate user (%s)" % self._user)
        return self._proxy[self._db]

    def set_collection(self, collection):
        self.collection = collection

    def __getitem__(self, collection):
        return self.db[collection]



#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward <cward@redhat.com>

import logging
logger = logging.getLogger(__name__)

from copy import copy
from datetime import datetime
import simplejson as json
import sys
import traceback

from metriqued.config import mongodb
from metriqued.defaults import MONGODB_CONF
from metriqued.utils import dt2ts, new_oid


class Job(object):
    def __init__(self, action, job_key=None):
        self._conf_mongodb = mongodb(MONGODB_CONF)
        self.c_job_activity = self._conf_mongodb.c_job_activity
        self.action = action
        self.objectid = job_key

        if job_key:
            # get previously init'd job...
            doc = self.c_job_activity.find_one(self._base_spec)
            if not doc:
                raise ValueError("Invalid job_key id: %s" % job_key)
            self.args = doc['args']
            self.created = doc['created']
            self.active = doc['active']
            self.error = doc['error']
        else:
            self.objectid = new_oid()
            self.created = dt2ts(datetime.utcnow())
            # which arguments are associated with this job
            self.args = None
            # will store datetime of completion, if there is one
            self.completed = None
            # default, we assume this job should run
            self.active = True
            # more defaults...
            self.error = None
            # initiate the job
            self.save()

    @property
    def _base_spec(self):
        return {'_id': str(self.objectid)}

    def stop(self):
        logger.debug('STOP: %s' % self.objectid)
        self.active = False
        self.save()

    @property
    def payload(self):
        now = datetime.utcnow()
        _payload = copy(self._base_spec)
        # certain content can have unpredictable keys names
        # that cause mongo issues if the data structures
        # are 'nested document' like so... dump as json
        _args = json.dumps(self.args, ensure_ascii=False)

        _payload.update({
            'atime': dt2ts(now),
            'created': dt2ts(self.created),
            'active': self.active,
            'completed': self.completed,
            'error': str(self.error),
            'class': str(self.__class__),
            'args': _args,
        })
        return _payload

    def save(self):
        try:
            self.c_job_activity.save(self.payload, safe=True)
        except Exception as e:
            # oops! well, log the issue and update the process job
            # in the db with the error so clients are aware of the
            # issue.
            tb = traceback.format_exc(sys.exc_info())
            logger.error(tb)
            raise RuntimeError(
                "Payload failed to save; dumping ERROR (%s)" % str(e))

    def complete(self):
        self.active = False
        now = datetime.utcnow()
        if not self.completed:
            # don't overwrite the completed datetime
            # but add it if we're completed for the first time
            self.completed = now
        self.save()


def get_job(action, job_key=None):
    return Job(action, job_key)


def job_save(name):
    def decorator(func):
        def wrapper(*args, **kwargs):
            job = get_job(name)
            logger.debug('Running: %s' % name)
            result = func(*args, **kwargs)
            job.complete()
            return result
        return wrapper
    return decorator
